*******************************************************************
**
**                     P R I N T 2 P D F
**
*******************************************************************
PROCEDURE Print2PDF
parameters pcFileName, pcReport

if vartype(pcFileName) <> "C" or vartype(pcReport) <> "C" or empty(pcFileName) or empty(pcReport)
	=messagebox("No Parms passed to Print2PDF",48,"Error")
	return .f.
endif

local loPDF
loPDF = .NULL.

loPDF = createobject("Print2PDF")

if isnull(loPDF)
	=messagebox("No puedo configurar PDF class!",48,"Error")
	return .f.
endif

with loPDF
	.cOutputFile 	= pcFileName
	.cReport		= pcReport
	llResult 		= .Main()
endwith

if !llResult or loPDF.lError
	=messagebox("Error al crear PDF:"+chr(13)+alltrim(loPDF.cError),48,"Error Message")
endif

loPDF = .NULL.
release loPDF

return .t.



**************************************************************************************
** Class Definition :: Print2PDF
**************************************************************************************
define class Print2PDF as relation

	**Set these properties (required)
	cReport			= space(0)	&&Name of VFP report to Run
	cOutputFile		= space(0)	&&Name of finished PDF file to create.

	**User-Definable properties (most of these can be set in the .ini file)
	cStartFolder	= justpath(sys(16))+iif(right(justpath(sys(16)),1)="\","","\")	&&Folder this program started from.
	cTempPath		= space(0)	&&Folder for Temporary Files (default = VFP temp path)
	cExtraRptClauses= space(0)	&&Any extra reporting clauses for the "report form" command
	lReadINI		= .t.		&&Do you want to pull settings out of Print2PDF.ini file?
	cINIFile		= this.cStartFolder+"Print2PDF.ini"	&&Name of INI file to use.  If not in current folder or VFP path, specify full path.
	lFoundPrinter	= .f.		&&Was the PS printer found?
	lFoundGS		= .f.		&&Was Ghostscript found?
	cPSPrinter		= space(0)	&&Name of the Windows Printer that is the Postscript Printer (default = "GENERIC POSTSCRIPT PRINTER")
	cPSColorPrinter	= space(0)	&&Name of the Windows Printer that is the Postscript Printer (default = "GENERIC COLOR POSTSCRIPT")
	lUseColor		= .f.		&&Use "color" printer driver?
	cPrintResolution= space(0)	&&Printer resolution string (300, 600x600, etc.) (default = "300")
	cPSFile			= space(0)	&&Path/Filename for Postscript file (auto-created if not passed)
	lErasePSFile	= .t.		&&Erase the .ps file after conversion?
	cPDFFile		= space(0)	&&Path/Filename for PDF file		(auto-created if not passed)
	cGSFolder		= space(0)	&&Path where Ghostscript DLLs exist	(auto-populated if not passed)


	**Internal properties
	lError			= .f.		&&Indicates that this class had an error.
	cError			= ""		&&Error message generated by this class
	cOrigSafety		= space(0)	&&Original "set safety" settting
	cOrigPrinter	= space(0)	&&Original "Set printer" setting


	**AutoInstall properties	&&See the ReadINI method for more details
	iInstallCount	= 1			&&Number of programs setup for AutoInstallation

	dimension aInstall[1, 7]

	aInstall[1,1] = space(0)	&&Program Identifier (used to find program in array)
	aInstall[1,2] = .t.			&&Can we install this product
	aInstall[1,3] = space(0)	&&Product Name (for user)
	aInstall[1,4] = space(0)	&&Description of product for user
	aInstall[1,5] = space(0)	&&Folder where install files are stored
	aInstall[1,6] = space(0)	&&Setup Executable name
	aInstall[1,7] = space(0)	&&Notes to show user before installing


	**************************************************************************************
	** Class Methods
	**************************************************************************************
	**********************
	** Init Method
	**********************
	function init(pcFileName, pcReport, plRunNow)
		with this
			.cOrigSafety = set("safety")
			.cOrigPrinter = set("Printer", 3)
			
			.cStartFolder = FULLPATH(".")                       && VES Jul 25, 2012
			.cINIFile = ADDBS(.cStartFolder) + "PRINT2PDF.INI"  && VES Jul 25, 2012

			.lError = .f.
			.cError = ""

			if type("pcFileName") = "C" and !empty(pcFileName)
				.cOutputFile = alltrim(pcFileName)
			endif

			if type("pcReport") = "C" and !empty(pcReport)
				.cReport = alltrim(pcReport)
			endif
		endwith

		set safety off

		**Did User pass in parm to autostart the Main method?
		if type("plRunNow") = "L" and plRunNow = .t.
			return this.main()
		endif
	endfunc


	**********************************************************************
	** Cleanup Method
	** 	Make sure all objects are released, etc.
	**********************************************************************
	function CleanUp
		local lcOrigPrinter, lcOrigSafety

		with this
			lcOrigSafety = .cOrigSafety
			lcOrigPrinter = .cOrigPrinter
		endwith

		if !empty(lcOrigSafety)
			set safety &lcOrigSafety
		endif

		if !empty(lcOrigPrinter)
			set printer to
			set printer to name "&lcOrigPrinter"
		endif

		return
	endfunc


	**********************************************************************
	** ResetError Method
	** 	Call this method on each subsequent call to SendFax or CheckLog
	**********************************************************************
	function ResetError
		with this
			.lError = .f.
			.iError = 0
			.cError = ""
		endwith
		return .t.
	endfunc



	**************************************************************************************
	* Main Method - Main code
	**************************************************************************************
	function main(pcFileName, pcReport)
		local x
		store 0 to x
	 	with this
			if type("pcFileName") = "C" and !empty(pcFileName)
				.cOutputFile = alltrim(pcFileName)
			endif

			if type("pcReport") = "C" and !empty(pcReport)
				.cReport = alltrim(pcReport)
			endif

			if empty(.cReport) or empty(.cOutputFile)
				.lError = .t.
				.cError(".cReport and/or .cOutputFile empty",48,"Error")
				return .f.
			endif

			**Get values from Print2Pdf.ini file
			**Also sets default values even if .ini is not used.
			if !.lError
				=.ReadINI()
			endif

			**Set printer to PostScript
			if !.lError
				=.SetPrinter()
			endif

			**Create the Postscript file
			if !.lError
				=.MakePS()
			endif

			**Make sure Ghostscript DLLs can be found
			if !.lError
				=.GSFind()
			endif

			**Turn Postscript into PDF
			if !.lError
				=.MakePDF()
			endif

			**Install PDF Reader
			if !.lError
				=.InstPDFReader()
			endif

			.CleanUp()
		endwith

		return !this.lError
	endfunc



	**************************************************************************************
	* ReadIni()	-	Function to open/read contents of Print2PDF.INI file.
	*			-	If the .ReadINI property is .False., this method will not run
	*			-	This method examines each property, it will not overwrite a property
	*				with a value from the .INI that you have already populated via code.
	**************************************************************************************
	function ReadINI()
		local lcTmp
		store "" to lcTmp

		**If we're not supposed to read the INI, make sure default values are set
		if this.lReadINI = .t.
			**Win API declaration
			declare integer GetPrivateProfileString ;
				in WIN32API ;
				string cSection,;
				string cEntry,;
				string cDefault,;
				string @cRetVal,;
				integer nSize,;
				string cFileName


			**Read INI settings
			with this
				**General Properties
				**Postscript Printer Driver Name
				if empty(.cPSPrinter)
					.cPSPrinter = .ReadIniSetting("PostScript", "cPSPrinter")
				endif

				**Color Postscript Printer Driver Name
				if empty(.cPSColorPrinter)
					.cPSColorPrinter = .ReadIniSetting("PostScript", "cPSColorPrinter")
				endif
		
				**Name of PostScript file
				if empty(.cPSFile)
					.cPSFile = .ReadIniSetting("PostScript", "cPSFile")
				endif

				**Name of folder to hold Temporary postscript files
				if empty(.cTempPath)
					.cTempPath = .ReadIniSetting("PostScript", "cTempPath")
				endif

				**Name of PDF file
				if empty(.cPDFFile)
					.cPDFFile = .ReadIniSetting("GhostScript", "cPDFFile")
				endif

				**Name of Ghostscript folder (where installed to)
				if empty(.cGSFolder)
					.cGSFolder = .ReadIniSetting("GhostScript", "cGSFolder")
				endif

				**Resolution for PDF files
				if empty(.cPrintResolution)
					.cPrintResolution = .ReadIniSetting("PostScript", "cPrintResolution")
				endif

				**Installation Packages
				**# of packages to store settings for
				lcTmp = .ReadIniSetting("Install", "iInstallCount")
				if !empty(lcTmp)
					.iInstallCount = val(lcTmp)

					if .iInstallCount > 1
						dimension .aInstall[.iInstallCount, 7]
					endif

					for x = 1 to .iInstallCount
						**What is the "programmatic" ID for this package
						.aInstall[x,1] = upper(.ReadIniSetting("Install", "cInstID"+transform(x)))

						**Can we install this package?
						lcTmp = upper(.ReadIniSetting("Install", "lAllowInst"+transform(x)))
						.aInstall[x,2] = iif("T" $ lcTmp or "Y" $ lcTmp, .t., .f.)

						**Product Name
						.aInstall[x,3] = .ReadIniSetting("Install", "cInstProduct"+transform(x))

						**Description of Product to show user
						.aInstall[x,4] = .ReadIniSetting("Install", "cInstUserDescr"+transform(x))

						**Folder where installation files exist
						.aInstall[x,5] = .ReadIniSetting("Install", "cInstFolder"+transform(x))

						**Executable file to start installation
						.aInstall[x,6] = .ReadIniSetting("Install", "cInstExe"+transform(x))
						
						**Instructions to User
						.aInstall[x,7] = .ReadIniSetting("Install", "cInstInstr"+transform(x))
					endfor
				endif
			ENDWITH
		ENDIF
		
		**Make sure these basic settings are not blank
		if empty(.cPSPrinter)
			.cPSPrinter	= "GENERIC POSTSCRIPT PRINTER"
		endif
		if empty(.cPSColorPrinter)
			.cPSColorPrinter = "GENERIC COLOR POSTSCRIPT"
		ENDIF
		if empty(.cTempPath)
			.cTempPath = sys(2023) + iif(right(sys(2023),1)="\","","\")
		endif
		if empty(.cPrintResolution)
			.cPrintResolution= "300"
		endif
		return .t.
	endfunc



	**************************************************************************************
	* ReadIniSetting() - Returns the value of a "setting" from an "INI" file (text file)
	*				 	 (returns "" if string is not found)
	*	Parms:	pcSection	= The "section" in the INI file to look in...	[Section Name]
	*			pcSetting	= The "setting" to return the value of			Setting="MySetting"
	**************************************************************************************
	function ReadIniSetting(pcSection, pcSetting)
		local lcRetValue, lnNumRet, lcFile
		
		lcFile = alltrim(this.cIniFile)

		lcRetValue = space(8196)

		**API call to get string
		lnNumRet = GetPrivateProfileString(pcSection, pcSetting, "[MISSING]", @lcRetValue, 8196, lcFile)

		
		lcRetValue = alltrim(substr(lcRetValue, 1, lnNumRet))
		
		if lcRetValue == "[MISSING]"
			lcRetValue = ""
		endif
		
		return lcRetValue
	endfunc



	**************************************************************************************
	* SetPrinter() - Set the printer to the PostScript Printer
	**************************************************************************************
	function SetPrinter()
		local x, lcPrinter
		x = 0
		lcPrinter = ""
		
		with this
			if empty(.cPSPrinter)
				.cPSPrinter = "GENERIC POSTSCRIPT PRINTER"
			endif
			if empty(.cPSColorPrinter)
				.cPSColorPrinter = "GENERIC COLOR POSTSCRIPT"
			endif

			if .lUseColor = .t.
				lcPrinter = .cPSColorPrinter
			else
				lcPrinter = .cPSPrinter
			ENDIF
			lcPrinter = UPPER(lcPrinter) && VES Jul 25, 2012
			
			.lFoundPrinter = .f.

			***Make sure a Postscript printer exists on this PC
			if aprinters(laPrinters) > 0
				for x = 1 to alen(laPrinters)
					if alltrim(upper(laPrinters[x])) == lcPrinter
						.lFoundPrinter = .t.
					endif
				endfor

				if !.lFoundPrinter
					.cError = lcPrinter+" is not installed!!"
					.lError = .t.
				endif
			else
				.cError = "NO printer drivers are installed!!"
				.lError = .t.
			endif

			if .lFoundPrinter
				*** Set the printer to Generic Postscript Printer
				lcEval = "SET PRINTER TO NAME '" +lcPrinter+"'"
				&lcEval

				if alltrim(upper(set("PRINTER",3))) == alltrim(upper(lcPrinter))
				else
					.cError = "Could not set printer to: "+alltrim(lcPrinter)
					.lError = .t.
					.lFoundPrinter = .f.
				endif
			endif

			**Auto-Install, If no PS printer was found.
			if !.lFoundPrinter
				if this.Install("POSTSCRIPT")	&&Install PS driver
					return this.SetPrinter()	&&Call this function again
				endif
			endif
		endwith

		return .lFoundPrinter
	endfunc



	**************************************************************************************
	* MakePS() - Run the VFP report to a PostScript file
	**************************************************************************************
	function MakePS()
		local lcReport, lcExtra, lcPSFile

		set safety off

		with this
			**If no PS printer was found yet, find it
			if !.lFoundPrinter
				if !.SetPrinter()
					return .f.
				endif
			endif

			lcReport	= .cReport
			lcExtra		= .cExtraRptClauses

			if empty(lcReport)
				.cError = "No Report File was specified."
				.lError = .t.
				return .f.
			endif

			if empty(.cPSFile)
				*** We'll create a Postscript Output File (use VFP temporary path and temp filename)
				.cPSFile = .cTempPath + sys(2015) + ".ps"
			endif

			lcPSFile = .cPSFile

			*** Make sure we erase existing file first
			erase (lcPSFile)

			report form (lcReport) &lcExtra noconsole to file &lcPSFile

			if !file(lcPSFile)
				.cError = "Could create PDF file"
				.lError = .t.
				return .f.
			endif
		endwith

		return .t.
	endfunc



	**************************************************************************************
	* GSFind() - Finds the Ghostscript DLL path and adds it to the VFP path
	**************************************************************************************
	function GSFind()
		local x, lcPath
		store "" to lcPath
		store 0 to x

		with this
			.lFoundGS = .f.

			**Look for Ghostscript DLL files.  If not in the VFP path, then GSFind().
			if file("gsdll32.dll")
				.lFoundGS = .t.
				return .t.
			endif

			**Try location specified in INI file
			if !empty(.cGSFolder)
				lcTmp = .cGSFolder + "gsdll32.dll"			&&Make sure the DLL file can be found
				if !file(lcTmp)
					.cGSFolder = ""
				endif
			endif

			*Look for them to exist in C:\gs\gsX.XX\bin\
			if empty(.cGSFolder)
				if !directory("C:\gs")
					return .f.
				endif

				liGS = adir(laGSFolders, "C:\gs\*.*","D")
				if liGS < 1
					return .f.
				endif

				for x = 1 to alen(laGSFolders,1)
					lcTmp = alltrim(upper(laGSFolders[x,1]))
					if "GS" = left(lcTmp,2) and "D" $ laGSFolders[x,5]
						.cGSFolder = lcTmp
						exit
					endif
				endfor

				if empty(.cGSFolder)
					return .f.
				endif

				.cGSFolder = "c:\gs\"+alltrim(.cGSFolder)+"\bin\"
			endif

			if !empty(.cGSFolder)
				lcTmp = .cGSFolder + "gsdll32.dll"			&&Make sure the DLL file can be found
				if !file(lcTmp)
					.cGSFolder = ""
				endif
			endif

			if empty(.cGSFolder)
				return .f.
			else
				.lFoundGS = .t.
			endif
		endwith

		lcPath = alltrim(set("Path"))
		set path to lcPath + ";" + .cGSFolder

		return .t.
	endfunc




	**************************************************************************************
	* MakePDF() - Run Ghostscript to create PDF file from the Postscript file
	**************************************************************************************
	function MakePDF()
		local lcPDFFile, lcOutputFile, lcPSFile

		set safety off

		with this
			**Make sure Ghostscript DLLs have been found (or install them)
			if !.lFoundGS
				if !.GSFind()

					**Auto-Install, Ghostscript
					if .Install("GHOSTSCRIPT")
						if !.GSFind()	&&Call function again
							.cError = "Could not Install Ghostscript!"
							.lError = .t.
							return .f.
						endif
					endif
				endif
			endif

			lcOutputFile	= .cOutputFile
			lcPSFile		= .cPSFile

			if empty(.cPDFFile)
				.cPDFFile = juststem(lcPSFile) + ".pdf"
			endif

			lcPDFFile = .cPDFFile
			erase (lcPDFFile)

			if !.GSConvertFile(lcPSFile, lcPDFFile)
				.cError = "Could not create: "+lcPDFFile
				.lError = .t.
			endif

			if !file(lcPDFFile)
				.cError = "Could not create: "+lcPDFFile
				.lError = .t.
			endif

			**Get rid of .ps file
			if .lErasePSFile
				erase (lcPSFile)
			endif

			**Make sure output file does not exist already
			erase (lcOutputFile)

			*** Move the temp file to the actual file name by renaming
			rename (lcPDFFile) to (lcOutputFile)

			if !file(lcOutputFile)
				.cError = "Could not rename file to "+lcOutputFile
				.lError = .t.
			endif
		endwith
		return !this.lError
	endfunc



	**************************************************************************************
	* GSConvert() - Sets up arguments that will be passed to Ghostscript DLL, calls GSCall
	**************************************************************************************
	function GSConvertFile(tcFileIn, tcFileOut)
		local lnGSInstanceHandle, lnCallerHandle, loHeap, lnElementCount, lcPtrArgs, lnCounter, lnReturn
		dimension  laArgs[11]

		store 0 to lnGSInstanceHandle, lnCallerHandle, lnElementCount, lnCounter, lnReturn
		store null to loHeap
		store "" to lcPtrArgs

		set safety off
		loHeap = createobject('Heap')

		**Declare Ghostscript DLLs
		
		declare long gsapi_new_instance in gsdll32.dll ;
			long @lngGSInstance, long lngCallerHandle
		declare long gsapi_delete_instance in gsdll32.dll ;
			long lngGSInstance
		declare long gsapi_init_with_args in gsdll32.dll ;
			long lngGSInstance, long lngArgumentCount, ;
			long lngArguments
		declare long gsapi_exit in gsdll32.dll ;
			long lngGSInstance

		laArgs[1] = "dummy" 			&&You could specify a text file here with commands in it (NOT USED)
		laArgs[2] = "-dNOPAUSE"			&&Disables Prompt and Pause after each page
		laArgs[3] = "-dBATCH"			&&Causes GS to exit after processing file(s)
		laArgs[4] = "-dSAFER"			&&Disables the ability to deletefile and renamefile externally
		laArgs[5] = "-r"+this.cPrintResolution	&&Printer Resolution (300x300, 360x180, 600x600, etc.)
		laArgs[6] = "-sDEVICE=pdfwrite"	&&Specifies which "Device" (output type) to use.  "pdfwrite" means PDF file.
		laArgs[7] = "-sOutputFile=" + tcFileOut	&&Name of the output file
		laArgs[8] = "-c"				&&Interprets arguments as PostScript code up to the next argument that begins with "-" followed by a non-digit, or with "@". For example, if the file quit.ps contains just the word "quit", then -c quit on the command line is equivalent to quit.ps there. Each argument must be exactly one token, as defined by the token operator
		laArgs[9] = ".setpdfwrite"		&&If this file exists, it uses it as command-line input?
		laArgs[10] = "-f"				&&(ends the -c argument started in laArgs[8])
		laArgs[11] = tcFileIn			&&Input File name (.ps file)

		* Load Ghostscript and get the instance handle
		lnReturn = gsapi_new_instance(@lnGSInstanceHandle, @lnCallerHandle)
		if (lnReturn < 0)
			loHeap = null
			RELEASE loHeap
			this.lError = .t.
			this.cError = "Could not start Ghostscript."
			return .f.
		endif

		* Convert the strings to null terminated ANSI byte arrays
		* then get pointers to the byte arrays.
		lnElementCount = alen(laArgs)
		lcPtrArgs = ""
		for lnCounter = 1 to lnElementCount
			lcPtrArgs = lcPtrArgs + NumToLONG(loHeap.AllocString(laArgs[lnCounter]))
		endfor
		lnPtr = loHeap.AllocBlob(lcPtrArgs)

		lnReturn = gsapi_init_with_args(lnGSInstanceHandle, lnElementCount, lnPtr)
		if (lnReturn < 0)
			loHeap = null
			RELEASE loHeap
			this.lError = .t.
			this.cError = "Could not Initilize Ghostscript."
			return .f.
		endif

		* Stop the Ghostscript interpreter
		lnReturn=gsapi_exit(lnGSInstanceHandle)
		if (lnReturn < 0)
			loHeap = null
			RELEASE loHeap
			this.lError = .t.
			this.cError = "Could not Exit Ghostscript."
			return .f.
		endif


		* release the Ghostscript instance handle'
		=gsapi_delete_instance(lnGSInstanceHandle)

		loHeap = null
		RELEASE loHeap

		if !file(tcFileOut)
			this.lError = .t.
			this.cError = "Ghostscript could not create the PDF."
			return .f.
		endif

		return .t.
	endfunc



	**************************************************************************************
	* InstPDFReader Method - Installs the PDFReader if needed
	**************************************************************************************
	function InstPDFReader()
		**Make sure the PDF file has been created
		if !file(.cOutputFile)
			return .f.
		endif

		**Ask Windows which EXE is associated with this file (extension)
		lcExe = .AssocExe(.cOutputFile)

		if empty(lcExe)
			**Install the PDF Reader
			return .Install("PDFREADER")
		else
			return .t.
		endif
	endfunc


	**************************************************************************************
	* AssocExe Method - Returns the Executable File associated with a file
	**************************************************************************************
	function AssocExe(pcFile)
		local lcExeFile
		store "" to lcExeFile

		declare integer FindExecutable in shell32;
			string   lpFile,;
			string   lpDirectory,;
			string @ lpResult

		lcExeFile = space(250)

		if FindExecutable(pcFile, "", @lcExeFile) > 32
			lcExeFile = left(lcExeFile, at(chr(0), lcExeFile) -1)
		else
			lcExeFile = ""
		endif

		return lcExeFile
	endfunc


	**************************************************************************************
	* Install Method - Installs software on the PC
	**************************************************************************************
	function Install(pcID)
		local llFound, x, lcEval, lcProduct, lcDesc, lcTmp, ;
				lcFolder, lcInstEXE, lcInstruct, llDynaPath
		store "" to lcEval, lcProduct, lcAbbr, lcDesc, lcTmp, lcFolder, lcInstEXE, lcInstruct
		store .f. to llFound, llDynaPath

		with this
			pcID = alltrim(upper(pcID))

			**See if this Installation ID is in our array
			for x = 1 to alen(.aInstall,1)
				if alltrim(upper(.aInstall[x,1])) == pcID
					llFound = .t.
					exit
				endif
			endfor

			if !llFound
				.lError = .t.
				.cError = "Installation parms do not exist for ID: "+pcID
				return .f.
			endif

			**Copy array contents to variables
			llDoInst	= .aInstall[x,2]
			lcProduct	= .aInstall[x,3]
			lcDesc		= .aInstall[x,4]
			lcFolder	= .aInstall[x,5]
			lcInstEXE	= .aInstall[x,6]
			if !empty(.aInstall[x,7])
				lcInstruct	= ALLTRIM(.aInstall[x,7])
				if "+" $ lcInstruct
					lcInstruct = &lcInstruct
				endif
			else
				lcInstruct	= "Please accept the 'Default Values'"+chr(13)+"during the installation."
			endif

			**See if the path is "dynamically" generated based on variables
			if "+" $ lcFolder
				llDynaPath = .t.
			else
				llDynaPath = .f.
			endif
			
			**Are we allowed to install this product?
			if llDoInst = .t.
				**Make sure we have the Folder and Executable to install from?
				if !empty(lcFolder) and !empty(lcInstEXE)
					if llDynaPath
						lcFolder = alltrim(lcFolder)
						lcEval = &lcFolder
						lcEval = lcEval+alltrim(lcInstEXE)	&&command string
					else
						if right(lcFolder,1) <> "\"				&&Make sure the final backslash exists
							lcFolder = lcFolder + "\"
						endif
				
						lcEval = alltrim(lcFolder)+alltrim(lcInstEXE)	&&command string
					endif

					**Make sure install .exe exists in the path given
					if !llDynaPath and !file(lcEval)
						.cError = "No puedo encontrar el instalador para "+lcProduct+" in:"+chr(13)+alltrim(lcEval)
						.lError = .t.
					else
						if 7=messagebox(lcProduct+" necesita estar instalado en tu computador."+chr(13)+;
								lcDesc+chr(13)+;
								"Instalar ahora?",36,"Confirme")
							.lError = .t.
							.cError = "Cancelado por el usuario "+lcProduct+" Instalacin."
							return .f.
						endif

						=messagebox(lcInstruct,64,"Instrucciones")

						**Do the Installation
						.aInstall[x,2] = .f.		&&Do not allow ourselves to get into a loop
						lcEval = "run /n "+lcEval
						&lcEval

						=messagebox("When the Installation has finished"+chr(13)+;
							"COMPLETELY, please click OK...",64,"Waiting for Installation...")

						**Did it work?
						if 7=messagebox("Was the installation successfull?"+chr(13)+chr(13)+;
								"If no errors ocurred during the Installation"+chr(13)+;
								"and everything went OK, please click 'Yes'...",36,"Everything OK?")
							.lError = .t.
							.cError = "Errors ocurred during "+lcProduct+" Installation."
							return .f.
						else
							.lError = .f.
							.cError = ""
							return .t.
						endif
					endif
				endif
			endif
		endwith
		return .f.
	endfunc

enddefine

*** End of Class Print2PDF ***




**************************************************
*-- Class:        heap 
*-- ParentClass:  custom
*-- BaseClass:    custom
*
*****

DEFINE CLASS heap AS custom

	PROTECTED inHandle, inNumAllocsActive,iaAllocs[1,3]
	inHandle = NULL
	inNumAllocsActive = 0
	iaAllocs = NULL
	Name = "heap"

	PROCEDURE Alloc
		*  Allocate a block, returning a pointer to it
		LPARAMETER nSize
		DECLARE INTEGER HeapAlloc IN WIN32API AS HAlloc;
			INTEGER hHeap, ;
			INTEGER dwFlags, ;
			INTEGER dwBytes
		DECLARE INTEGER HeapSize IN WIN32API AS HSize ;
			INTEGER hHeap, ;
			INTEGER dwFlags, ;
			INTEGER lpcMem
		LOCAL nPtr
		WITH this
			nPtr = HAlloc(.inHandle, 0, @nSize)
			IF nPtr # 0
				*  Bump the allocation array
				.inNumAllocsActive = .inNumAllocsActive + 1
				DIMENSION .iaAllocs[.inNumAllocsActive,3]
				*  Pointer
				.iaAllocs[.inNumAllocsActive,1] = nPtr
				*  Size actually allocated - get with HeapSize()
				.iaAllocs[.inNumAllocsActive,2] = HSize(.inHandle, 0, nPtr)
				*  It's alive...alive I tell you!
				.iaAllocs[.inNumAllocsActive,3] = .T.
			ELSE
				*  HeapAlloc() failed - return a NULL for the pointer
				nPtr = NULL
			ENDIF
		ENDWITH
		RETURN nPtr
	ENDPROC

	FUNCTION AllocBLOB
		LPARAMETER cBStringToCopy
		LOCAL nAllocPtr
		WITH this
			nAllocPtr = .Alloc(LEN(cBStringToCopy))
			IF ! ISNULL(nAllocPtr)
				.CopyTo(nAllocPtr,cBStringToCopy)
			ENDIF
		ENDWITH
		RETURN nAllocPtr
	ENDFUNC
	
	FUNCTION AllocString
		LPARAMETER cString
		RETURN this.AllocBLOB(cString + CHR(0))
	ENDFUNC
	
	FUNCTION AllocInitAs
		LPARAMETER nSizeOfBuffer, nByteValue
		IF TYPE('nByteValue') # 'N' OR ! BETWEEN(nByteValue,0,255)
			nByteValue = 0
		ENDIF
		RETURN this.AllocBLOB(REPLICATE(CHR(nByteValue),nSizeOfBuffer))
	ENDFUNC

	PROCEDURE DeAlloc
		LPARAMETER nPtr
		DECLARE INTEGER HeapFree IN WIN32API AS HFree ;
			INTEGER hHeap, ;
			INTEGER dwFlags, ;
			INTEGER lpMem
		LOCAL nCtr
		nCtr = NULL
		WITH this
			nCtr = .FindAllocID(nPtr)
			IF ! ISNULL(nCtr)
				=HFree(.inHandle, 0, nPtr)
				.iaAllocs[nCtr,3] = .F.
			ENDIF
		ENDWITH
		RETURN ! ISNULL(nCtr)
	ENDPROC

	PROCEDURE CopyTo
		LPARAMETER nPtr, cSource
		DECLARE RtlMoveMemory IN WIN32API AS RtlCopy ;
			INTEGER nDestBuffer, ;
			STRING @pVoidSource, ;
			INTEGER nLength
		LOCAL nCtr
		nCtr = NULL
		IF TYPE('nPtr') = 'N' AND TYPE('cSource') $ 'CM' ;
		   AND ! (ISNULL(nPtr) OR ISNULL(cSource))
			WITH this
				nCtr = .FindAllocID(nPtr)
				IF ! ISNULL(nCtr)
					=RtlCopy((.iaAllocs[nCtr,1]), ;
							cSource, ;
							MIN(LEN(cSource),.iaAllocs[nCtr,2]))
				ENDIF
			ENDWITH
		ENDIF
		RETURN ! ISNULL(nCtr)
	ENDPROC


	PROCEDURE CopyFrom
		LPARAMETER nPtr
		DECLARE RtlMoveMemory IN WIN32API AS RtlCopy ;
			STRING @DestBuffer, ;
			INTEGER pVoidSource, ;
			INTEGER nLength
		LOCAL nCtr, uBuffer
		uBuffer = NULL
		nCtr = NULL
		IF TYPE('nPtr') = 'N' AND ! ISNULL(nPtr)
			WITH this
				nCtr = .FindAllocID(nPtr)
				IF ! ISNULL(nCtr)
					uBuffer = REPL(CHR(0),.iaAllocs[nCtr,2])
					=RtlCopy(@uBuffer, ;
							(.iaAllocs[nCtr,1]), ;
							(.iaAllocs[nCtr,2]))
				ENDIF
			ENDWITH
		ENDIF
		RETURN uBuffer
	ENDPROC
	
	PROTECTED FUNCTION FindAllocID
	 	LPARAMETER nPtr
	 	LOCAL nCtr
	 	WITH this
			FOR nCtr = 1 TO .inNumAllocsActive
				IF .iaAllocs[nCtr,1] = nPtr AND .iaAllocs[nCtr,3]
					EXIT
				ENDIF
			ENDFOR
			RETURN IIF(nCtr <= .inNumAllocsActive,nCtr,NULL)
		ENDWITH
	ENDPROC

	PROCEDURE SizeOfBlock
		LPARAMETERS nPtr
		LOCAL nCtr, nSizeOfBlock
		nSizeOfBlock = NULL
		WITH this
			nCtr = .FindAllocID(nPtr)
			RETURN IIF(ISNULL(nCtr),NULL,.iaAllocs[nCtr,2])
		ENDWITH
	ENDPROC

	PROCEDURE Destroy
		DECLARE HeapDestroy IN WIN32API AS HDestroy ;
		  INTEGER hHeap

		LOCAL nCtr
		WITH this
			FOR nCtr = 1 TO .inNumAllocsActive
				IF .iaAllocs[nCtr,3]
					.Dealloc(.iaAllocs[nCtr,1])
				ENDIF
			ENDFOR
			HDestroy[.inHandle]
		ENDWITH
		DODEFAULT()
	ENDPROC


	PROCEDURE Init
		DECLARE INTEGER HeapCreate IN WIN32API AS HCreate ;
			INTEGER dwOptions, ;
			INTEGER dwInitialSize, ;
			INTEGER dwMaxSize
		#DEFINE SwapFilePageSize  4096
		#DEFINE BlockAllocSize    2 * SwapFilePageSize
		WITH this
			.inHandle = HCreate(0, BlockAllocSize, 0)
			DIMENSION .iaAllocs[1,3]
			.iaAllocs[1,1] = 0
			.iaAllocs[1,2] = 0
			.iaAllocs[1,3] = .F.
			.inNumAllocsActive = 0
		ENDWITH
		RETURN (this.inHandle # 0)
	ENDPROC


ENDDEFINE
*
*-- EndDefine: heap
**************************************************
*
*  Additional functions for working with structures and pointers and stuff
*
FUNCTION SetMem
LPARAMETERS nPtr, cSource
DECLARE RtlMoveMemory IN WIN32API AS RtlCopy ;
	INTEGER nDestBuffer, ;
	STRING @pVoidSource, ;
	INTEGER nLength

RtlCopy(nPtr, ;
		cSource, ;
		LEN(cSource))
RETURN .T.

FUNCTION GetMem
LPARAMETERS nPtr, nLen
DECLARE RtlMoveMemory IN WIN32API AS RtlCopy ;
	STRING @DestBuffer, ;
	INTEGER pVoidSource, ;
	INTEGER nLength
LOCAL uBuffer
uBuffer = REPL(CHR(0),nLen)
=RtlCopy(@uBuffer, ;
		 nPtr, ;
		 nLen)
RETURN uBuffer

FUNCTION GetMemString
LPARAMETERS nPtr, nSize
DECLARE INTEGER lstrcpyn IN WIN32API AS StrCpyN ;
	STRING @ lpDestString, ;
	INTEGER lpSource, ;
	INTEGER nMaxLength
LOCAL uBuffer
IF TYPE('nSize') # 'N' OR ISNULL(nSize)
	nSize = 512
ENDIF
uBuffer = REPL(CHR(0), nSize)
IF StrCpyN(@uBuffer, nPtr, nSize-1) # 0
	uBuffer = LEFT(uBuffer, MAX(0,AT(CHR(0),uBuffer) - 1))
ELSE
	uBuffer = NULL
ENDIF
RETURN uBuffer

FUNCTION SHORTToNum
 	LPARAMETER tcInt
	LOCAL b0,b1,nRetVal
	b0=asc(tcInt)
	b1=asc(subs(tcInt,2,1))
	if b1<128
		nRetVal=b1 * 256 + b0
	else
		b1=255-b1
		b0=256-b0
		nRetVal= -( (b1 * 256) + b0)
	endif
	return nRetVal

FUNCTION NumToSHORT
	LPARAMETER tnNum
	LOCAL b0,b1,x
	IF tnNum>=0
		x=INT(tnNum)
		b1=INT(x/256)
		b0=MOD(x,256)
	ELSE
		x=INT(-tnNum)
		b1=255-INT(x/256)
		b0=256-MOD(x,256)
		IF b0=256
			b0=0
			b1=b1+1
		ENDIF
	ENDIF
	RETURN CHR(b0)+CHR(b1)

FUNCTION DWORDToNum
	LPARAMETER tcDWORD
	LOCAL b0,b1,b2,b3
	b0=asc(tcDWORD)
	b1=asc(subs(tcDWORD,2,1))
	b2=asc(subs(tcDWORD,3,1))
	b3=asc(subs(tcDWORD,4,1))
	RETURN ( ( (b3 * 256 + b2) * 256 + b1) * 256 + b0)

FUNCTION NumToDWORD
	LPARAMETER tnNum
	RETURN NumToLong(tnNum)

FUNCTION WORDToNum
	LPARAMETER tcWORD
	RETURN (256 *  ASC(SUBST(tcWORD,2,1)) ) + ASC(tcWORD)

FUNCTION NumToWORD
	LPARAMETER tnNum
	LOCAL x
	x=INT(tnNum)
	RETURN CHR(MOD(x,256))+CHR(INT(x/256))
	
FUNCTION NumToLong
	LPARAMETER tnNum
	DECLARE RtlMoveMemory IN WIN32API AS RtlCopyLong ;
		STRING @pDestString, ;
		INTEGER @pVoidSource, ;
		INTEGER nLength
	LOCAL cString
	cString = SPACE(4)
   =RtlCopyLong(@cString, BITOR(tnNum,0), 4)
	RETURN cString
	
FUNCTION LongToNum
	LPARAMETER tcLong
	DECLARE RtlMoveMemory IN WIN32API AS RtlCopyLong ;
		INTEGER @ DestNum, ;
		STRING @ pVoidSource, ;
		INTEGER nLength
	LOCAL nNum
	nNum = 0
	=RtlCopyLong(@nNum, tcLong, 4)
	RETURN nNum
	
FUNCTION AllocNetAPIBuffer
LPARAMETER nSize
IF TYPE('nSize') # 'N' OR nSize <= 0
	RETURN NULL
ENDIF
IF ! 'NT' $ OS()
	RETURN NULL
ENDIF
DECLARE INTEGER NetApiBufferAllocate IN NETAPI32.DLL ;
	INTEGER dwByteCount, ;
	INTEGER lpBuffer
LOCAL  nBufferPointer
nBufferPointer = 0
IF NetApiBufferAllocate(INT(nSize), @nBufferPointer) # 0
	nBufferPointer = NULL
ENDIF
RETURN nBufferPointer

FUNCTION DeAllocNetAPIBuffer
LPARAMETER nPtr
IF TYPE('nPtr') # 'N'
	RETURN .F.
ENDIF
IF ! 'NT' $ OS()
	RETURN .F.
ENDIF
DECLARE INTEGER NetApiBufferFree IN NETAPI32.DLL ;
	INTEGER lpBuffer
RETURN (NetApiBufferFree(INT(nPtr)) = 0)

Function CopyDoubleToString
LPARAMETER nDoubleToCopy
DECLARE RtlMoveMemory IN WIN32API AS RtlCopyDbl ;
	STRING @DestString, ;
	DOUBLE @pVoidSource, ;
	INTEGER nLength
LOCAL cString
cString = SPACE(8)
=RtlCopyDbl(@cString, nDoubleToCopy, 8)
RETURN cString

FUNCTION DoubleToNum
LPARAMETER cDoubleInString
DECLARE RtlMoveMemory IN WIN32API AS RtlCopyDbl ;
	DOUBLE @DestNumeric, ;
	STRING @pVoidSource, ;
	INTEGER nLength
LOCAL nNum
nNum = 0.000000000000000000
=RtlCopyDbl(@nNum, cDoubleInString, 8)
RETURN nNum

*** End of CLSHEAP ***